/****************************************************************************************************************************************************
 * Copyright 2020 NXP
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *    * Redistributions in binary form must reproduce the above copyright notice,
 *      this list of conditions and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 *    * Neither the name of the NXP. nor the names of
 *      its contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************************************************************************************/

#include <FslBase/Math/Pixel/TypeConverter.hpp>
#include <FslGraphics/Font/BitmapFont.hpp>
#include <FslGraphics/Sprite/Font/SpriteFontFastLookup.hpp>
#include <FslGraphics/Sprite/SpriteNativeAreaCalc.hpp>
#include <FslGraphics/Sprite/SpriteUnitConverter.hpp>
#include <cassert>
#include <utility>
#include <vector>

namespace Fsl
{
  SpriteFontFastLookup::SpriteFontFastLookup(const SpriteNativeAreaCalc& spriteNativeAreaCalc, const PxExtent2D textureExtentPx,
                                             const BitmapFont& bitmapFont, const uint32_t densityDpi)
    : m_charLookup(Populate(spriteNativeAreaCalc, textureExtentPx, bitmapFont.GetChars(), bitmapFont.GetDpi(), densityDpi))
    , m_kerningLookup(bitmapFont.GetKernings())
    , m_lineSpacingPx(bitmapFont.GetLineSpacingPx())
    , m_baseLinePx(bitmapFont.GetBaseLinePx())
    , m_fontType(bitmapFont.GetFontType())
  {
  }

  BasicFlatHashTable<SpriteFontCharInfo, uint32_t, 256u> SpriteFontFastLookup::Populate(const SpriteNativeAreaCalc& spriteNativeAreaCalc,
                                                                                        const PxExtent2D textureExtentPx,
                                                                                        const ReadOnlySpan<BitmapFontChar> srcSpan,
                                                                                        const uint32_t imageDpi, const uint32_t densityDpi)
  {
    SpriteUnitConverter unitConverter(densityDpi);
    //

    const std::size_t count = srcSpan.size();
    std::vector<SpriteFontCharInfo> dst(count);
    for (std::size_t i = 0; i < count; ++i)
    {
      const BitmapFontChar& src = srcSpan[i];

      NativeTextureArea textureArea = spriteNativeAreaCalc.CalcNativeTextureArea(src.SrcTextureRectPx, textureExtentPx);
      PxSize2DF scaledSizePxf = unitConverter.CalcScaledPxSize2DF(src.SrcTextureRectPx.GetExtent(), imageDpi);
      PxVector2 offsetPxf = unitConverter.CalcScaledOffsetPxVector2(src.OffsetPx, imageDpi);
      float xAdvancePxf = unitConverter.CalcScaledOffsetValuePxf(src.XAdvancePx, imageDpi);

      dst[i] = SpriteFontCharInfo(CoreFontCharInfo(src.Id, TypeConverter::To<PxRectangleU16>(src.SrcTextureRectPx), src.OffsetPx, src.XAdvancePx),
                                  RenderFontCharInfo(textureArea, scaledSizePxf, offsetPxf, xAdvancePxf));
    }
    return BasicFlatHashTable<SpriteFontCharInfo, uint32_t, 256u>(ReadOnlySpanUtil::AsSpan(dst));
  }
}
